/*
 * Copyright 2008  Rice University
 * All Rights Reserved
 * 
 *       Created by Erion Plaku, Lydia Kavraki
 *       Physical and Biological Computing Group
 *       Department of Computer Science
 *       Rice University
 */

/**
 *@file OOPSMPRRT.H
 *@author Erion Plaku 
 *@date First version: March 15, 2007.  Last update: March 15, 2007.
 *@brief Implementation of the Rapidly-exploring Random Tree (RRT)
 *       motion planner  
 *
 *@remark
 * Algorithm described in Rapidly-exploring random trees: Progress and
 * prospects. S. M. LaValle and J. J. Kuffner. In B. R. Donald,
 * K. M. Lynch, and D. Rus, editors, Algorithmic and Computational
 * Robotics: New Directions, pages 293--308. A K Peters, Wellesley, MA,
 * 2001 
 */

#ifndef OOPSMP_MOTION_PLANNERS_TREE_PLANNER_RRT__OOPSMP_RRT_H_
#define OOPSMP_MOTION_PLANNERS_TREE_PLANNER_RRT__OOPSMP_RRT_H_

#include "MotionPlanners/TreePlanner/OOPSMPTreePlanner.H"
#include "Utils/DataStructures/Proximity/OOPSMPProximity.H"

namespace OOPSMP
{
    /// @cond
    /**
     *@author Erion Plaku
     *@brief Macro for class, class pointer, and const class pointer
     *       declarations 
     */  
    ForwardClassDeclaration(RRT);
    /// @endcond

   /**
    *@author Erion Plaku
    *@brief Implementation of the Rapidly-exploring Random Tree (RRT)
    *       motion planner   
    *
    *@remark
    * Algorithm described in Rapidly-exploring random trees: Progress and
    * prospects. S. M. LaValle and J. J. Kuffner. In B. R. Donald,
    * K. M. Lynch, and D. Rus, editors, Algorithmic and Computational
    * Robotics: New Directions, pages 293--308. A K Peters, Wellesley, MA,
    * 2001 
    *
    *@par Plug-and-play functionality:
    *  @code
    *  BeginImplementInstanceFactory(RRT, TreePlannerFactory);
    *    RegisterFnFactory(setProximity,   FACTORY_PTR_ARG(ProximityKeyPointer_t, 0));
    *  EndImplementInstanceFactory(RRT);
    *  @endcode 
    */  
    class RRT : public TreePlanner
    {
    public:
	/**
	 *@author Erion Plaku
	 *@brief Initialize data elements to appropriate default
	 *       values 
         */
	RRT(void);

	/**
	 *@author Erion Plaku
	 *@brief Free the memory allocated to different data elements 
         */	
	virtual ~RRT(void);

	/**
	 *@name Access motion planner data
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Get the proximity data structure that is used for
	 *       neighbor computations  
         */
	ProximityKeyPointer_t getProximity(void) const
	{
	    return m_proximity;
	}

	/**
	 *END Access motion planner data
	 *@}
	 */

        /**
	 *@name Setup motion planner
	 *@{
	 */

	
	/**
	 *@author Erion Plaku
	 *@brief Set the proximity data structure that should be used
	 *       for neighbor computations 
	 *@param proximity pointer to a proximity data structure 
	 *
	 *@remarks
	 *  The memory associated with <em>proximity</em> should not
	 *  be freed by the calling function. This memory will be
	 *  freed by the destructor of RRT. 
         */
	void setProximity(ProximityKeyPointer_t proximity)
	{
	    m_proximity = proximity;
	}

        /**
	 *END Setup motion planner
	 *@}
	 */

        /**
	 *@name How to solve the current motion planner query
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Clear the data structures and other data elements in
	 *       preparation for solving motion planning queries 
         */
	void clear(void);

	/**
	 *@author Erion Plaku
	 *@brief Initialize or construct data structures that will
	 *       facilitate the solution of motion planning queries 
	 *@param ptime maximum amount of preprocessing time allocated
	 *       to motion planner 
 	 *@par Description
	 *   Motion planner uses at most <em>ptime</em> seconds for
	 *   preprocessing.  Tree planner uses the preprocessing time
	 *   to initialize the tree for the current query.  
	 */
	void preprocess(const double ptime);

	/**
	 *END How to solve the current motion planner query
	 *@}
	 */

	/**
	 *@name Grow tree
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Grow the tree rooted at the initial state toward the
	 *       goal state 
	 *
	 *@par Description
	 *  An iteration consists of generating a sample <em>x</em>,
	 *  finding several neighbors of <em>x</em> from the current
	 *  states of the tree, and then extending each neighbor
	 *  toward <em>x</em>. 
	 *
	 *@remark
	 *  This function is iteratively called from the function
	 *  TreePlanner::solve 
         */
	void iterate(void);

        /**
	 *@author Erion Plaku
	 *@brief If possible, grow the tree using the indicated state
	 *       as a suggestion for the growth direction  
	 *
	 *@param x state
	 *
	 *@returns
	 *  id of the new state that is added to the tree or a
	 *  negative number if no states are added to the tree
	 */
	int extendTowardState(State_const_t x);

	/**
	 *END Grow tree
	 *@}
	 */

	/**
	 *@name Add states/root to tree
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Add a new vertex and a new branch to the tree 
	 *
	 *@param idFrom index of vertex
	 *@param xTo ending state
 	 *@param pathFromTo path connecting vertex with index
	 *       <em>idFrom</em> to the vertex <em>xTo</em> 
	 *
	 *@par Description
	 *  Add a new vertex to the tree and associate <em>xTo</em>
	 *  with the new vertex data. Add a branch to the tree,
	 *  connecting the vertex with index <em>idFrom</em> to the
	 *  new vertex and associate the path <em>pathFromTo</em> with
	 *  the edge data. 
         */
	int addState(const int idFrom, State_t xTo, Path_t pathFromTo)
	{
	    const int id = TreePlanner::addState(idFrom, xTo, pathFromTo);
	    m_proximity->addKey(getLastVertexData());
	    return id;				
	}

	/**
	 *END Add states/root to tree
	 *@}
	 */

    protected:
	/**
	 *@name Distance function for proximity computations
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Distance function used for neighbor computations
	 *@param key1 index of the first vertex
	 *@param key2 index of the second vertex
	 *@param keyDistFnData pointer to an object of type StateSpace 
	 *
	 *@par Description
	 *  The StateSpace::distanceStates function is used for
	 *  neighbor computations 
         */
	static double proximityKeyDistFn(const OOPSMPpointer_t key1, const OOPSMPpointer_t key2, OOPSMPpointer_t keyDistFnData)
	{
	    return ((StateSpace_const_t) keyDistFnData)->distanceStates(((TreePlanner::VertexData_t) key1)->getState(),
									((TreePlanner::VertexData_t) key2)->getState());
	}

	/**
	 *END Distance function for proximity computations
	 *@}
	 */

    protected:
	/**
	 *@name Variables for proximity
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Data structure for computing neighbors
	 */
	ProximityKeyPointer_t           m_proximity;

        /**
	 *@author Erion Plaku
	 *@brief Data structure for the current proximity query
	 */
	ProximityQuery<OOPSMPpointer_t> m_proximityQuery;


        /**
	 *@author Erion Plaku
	 *@brief Data structure for the results of the neighbor
	 *       computations 
	 */
	ProximityResults<OOPSMPpointer_t>  m_proximityResults;

	/**
	 *@author Erion Plaku
	 *@brief Auxiliary vertex data used for neighbors computations 
	 */
	TreePlanner::VertexData_t  m_aux_vReach;

	/**
	 *END Variables for proximity
	 *@}
	 */

	/**
	 *@name Variables for tree growth
	 *@{
	 */

	/**
	 *@author Erion Plaku
	 *@brief Auxiliary variable to store the random state used to
	 *       specify the direction of the growth 
	 */
	State_t                         m_aux_xRand;

	/**
	 *@author Erion Plaku
	 *@brief Auxiliary variable to store the new state added to
	 *       the tree 
	 */
	State_t                         m_aux_xNew;

	/**
	 *END Variables for tree growth
	 *@}
	 */
    };
    
    /**
     *@author Erion Plaku
     *@brief Declare factory and specify in source file the functions
     *       that should be registered for plug-and-play functionality 
     */     
     DeclareInstanceFactory(RRT, TreePlannerFactory);
}

#endif








